@@ -202,10 +202,13 @@ const char *px_strerror(int err);
 const char *px_resolve_alias(const PX_Alias *aliases, const char *name);
 
 void		px_set_debug_handler(void (*handler) (const char *));
+
 void		px_memset(void *ptr, int c, size_t len);
 
 #ifdef PX_DEBUG
-void		px_debug(const char *fmt,...);
+void
+px_debug(const char *fmt,...)
+__attribute__((format(PG_PRINTF_ATTRIBUTE, 1, 2)));
 #else
 #define px_debug(...)
 #endif
