@@ -2,7 +2,7 @@
  * pltcl.c		- PostgreSQL support for Tcl as
  *				  procedural language (PL)
  *
- *	  $PostgreSQL: pgsql/src/pl/tcl/pltcl.c,v 1.134 2010/07/06 19:19:01 momjian Exp $
+ *	  src/pl/tcl/pltcl.c
  *
  **********************************************************************/
 
@@ -18,6 +18,12 @@
 #define CONST84
 #endif
 
+/* ... and for Tcl 8.6. */
+#ifndef CONST86
+#define CONST86
+#endif
+
+#include "access/htup_details.h"
 #include "access/xact.h"
 #include "catalog/pg_proc.h"
 #include "catalog/pg_type.h"
@@ -31,6 +37,7 @@
 #include "utils/builtins.h"
 #include "utils/lsyscache.h"
 #include "utils/memutils.h"
+#include "utils/rel.h"
 #include "utils/syscache.h"
 #include "utils/typcache.h"
 
@@ -95,9 +102,9 @@ PG_MODULE_MAGIC;
  **********************************************************************/
 typedef struct pltcl_interp_desc
 {
-	Oid			user_id;				/* Hash key (must be first!) */
-	Tcl_Interp *interp;					/* The interpreter */
-	Tcl_HashTable query_hash;			/* pltcl_query_desc structs */
+	Oid			user_id;		/* Hash key (must be first!) */
+	Tcl_Interp *interp;			/* The interpreter */
+	Tcl_HashTable query_hash;	/* pltcl_query_desc structs */
 } pltcl_interp_desc;
 
 
@@ -127,7 +134,7 @@ typedef struct pltcl_proc_desc
 typedef struct pltcl_query_desc
 {
 	char		qname[20];
-	void	   *plan;
+	SPIPlanPtr	plan;
 	int			nargs;
 	Oid		   *argtypes;
 	FmgrInfo   *arginfuncs;
@@ -137,7 +144,7 @@ typedef struct pltcl_query_desc
 
 /**********************************************************************
  * For speedy lookup, we maintain a hash table mapping from
- * function OID + trigger OID + user OID to pltcl_proc_desc pointers.
+ * function OID + trigger flag + user OID to pltcl_proc_desc pointers.
  * The reason the pltcl_proc_desc struct isn't directly part of the hash
  * entry is to simplify recovery from errors during compile_pltcl_function.
  *
@@ -148,14 +155,19 @@ typedef struct pltcl_query_desc
  **********************************************************************/
 typedef struct pltcl_proc_key
 {
-	Oid			proc_id;				/* Function OID */
-	Oid			trig_id;				/* Trigger OID, or 0 if not trigger */
-	Oid			user_id;				/* User calling the function, or 0 */
+	Oid			proc_id;		/* Function OID */
+
+	/*
+	 * is_trigger is really a bool, but declare as Oid to ensure this struct
+	 * contains no padding
+	 */
+	Oid			is_trigger;		/* is it a trigger function? */
+	Oid			user_id;		/* User calling the function, or 0 */
 } pltcl_proc_key;
 
 typedef struct pltcl_proc_ptr
 {
-	pltcl_proc_key proc_key;			/* Hash key (must be first!) */
+	pltcl_proc_key proc_key;	/* Hash key (must be first!) */
 	pltcl_proc_desc *proc_ptr;
 } pltcl_proc_ptr;
 
@@ -192,7 +204,7 @@ static HeapTuple pltcl_trigger_handler(PG_FUNCTION_ARGS, bool pltrusted);
 static void throw_tcl_error(Tcl_Interp *interp, const char *proname);
 
 static pltcl_proc_desc *compile_pltcl_function(Oid fn_oid, Oid tgreloid,
-											   bool pltrusted);
+					   bool pltrusted);
 
 static int pltcl_elog(ClientData cdata, Tcl_Interp *interp,
 		   int argc, CONST84 char *argv[]);
@@ -252,7 +264,7 @@ pltcl_FinalizeNotifier(ClientData clientData)
 }
 
 static void
-pltcl_SetTimer(Tcl_Time *timePtr)
+pltcl_SetTimer(CONST86 Tcl_Time *timePtr)
 {
 }
 
@@ -278,7 +290,7 @@ pltcl_ServiceModeHook(int mode)
 }
 
 static int
-pltcl_WaitForEvent(Tcl_Time *timePtr)
+pltcl_WaitForEvent(CONST86 Tcl_Time *timePtr)
 {
 	return 0;
 }
@@ -488,12 +500,13 @@ pltcl_init_load_unknown(Tcl_Interp *interp)
 	 * This is for backwards compatibility.  To ensure that the table
 	 * is trustworthy, we require it to be owned by a superuser.
 	 ************************************************************/
-	pmrel = try_relation_openrv(makeRangeVar(NULL, "pltcl_modules", -1),
-								AccessShareLock);
+	pmrel = relation_openrv_extended(makeRangeVar(NULL, "pltcl_modules", -1),
+									 AccessShareLock, true);
 	if (pmrel == NULL)
 		return;
-	/* must be table or view, else ignore */
+	/* sanity-check the relation kind */
 	if (!(pmrel->rd_rel->relkind == RELKIND_RELATION ||
+		  pmrel->rd_rel->relkind == RELKIND_MATVIEW ||
 		  pmrel->rd_rel->relkind == RELKIND_VIEW))
 	{
 		relation_close(pmrel, AccessShareLock);
@@ -889,6 +902,8 @@ pltcl_trigger_handler(PG_FUNCTION_ARGS, bool pltrusted)
 			Tcl_DStringAppendElement(&tcl_cmd, "BEFORE");
 		else if (TRIGGER_FIRED_AFTER(trigdata->tg_event))
 			Tcl_DStringAppendElement(&tcl_cmd, "AFTER");
+		else if (TRIGGER_FIRED_INSTEAD(trigdata->tg_event))
+			Tcl_DStringAppendElement(&tcl_cmd, "INSTEAD OF");
 		else
 			elog(ERROR, "unrecognized WHEN tg_event: %u", trigdata->tg_event);
 
@@ -1046,18 +1061,20 @@ pltcl_trigger_handler(PG_FUNCTION_ARGS, bool pltrusted)
 			Oid			typioparam;
 			FmgrInfo	finfo;
 
-			/************************************************************
-			 * Ignore ".tupno" pseudo elements (see pltcl_set_tuple_values)
-			 ************************************************************/
-			if (strcmp(ret_name, ".tupno") == 0)
-				continue;
-
 			/************************************************************
 			 * Get the attribute number
+			 *
+			 * We silently ignore ".tupno", if it's present but doesn't match
+			 * any actual output column.  This allows direct use of a row
+			 * returned by pltcl_set_tuple_values().
 			 ************************************************************/
 			attnum = SPI_fnumber(tupdesc, ret_name);
 			if (attnum == SPI_ERROR_NOATTRIBUTE)
+			{
+				if (strcmp(ret_name, ".tupno") == 0)
+					continue;
 				elog(ERROR, "invalid attribute \"%s\"", ret_name);
+			}
 			if (attnum <= 0)
 				elog(ERROR, "cannot set system attribute \"%s\"", ret_name);
 
@@ -1170,7 +1187,7 @@ compile_pltcl_function(Oid fn_oid, Oid tgreloid, bool pltrusted)
 
 	/* Try to find function in pltcl_proc_htab */
 	proc_key.proc_id = fn_oid;
-	proc_key.trig_id = tgreloid;
+	proc_key.is_trigger = OidIsValid(tgreloid);
 	proc_key.user_id = pltrusted ? GetUserId() : InvalidOid;
 
 	proc_ptr = hash_search(pltcl_proc_htab, &proc_key,
@@ -1226,14 +1243,16 @@ compile_pltcl_function(Oid fn_oid, Oid tgreloid, bool pltrusted)
 		int			tcl_rc;
 
 		/************************************************************
-		 * Build our internal proc name from the functions Oid + trigger Oid
+		 * Build our internal proc name from the function's Oid.  Append
+		 * "_trigger" when appropriate to ensure the normal and trigger
+		 * cases are kept separate.
 		 ************************************************************/
 		if (!is_trigger)
 			snprintf(internal_proname, sizeof(internal_proname),
 					 "__PLTcl_proc_%u", fn_oid);
 		else
 			snprintf(internal_proname, sizeof(internal_proname),
-					 "__PLTcl_proc_%u_trigger_%u", fn_oid, tgreloid);
+					 "__PLTcl_proc_%u_trigger", fn_oid);
 
 		/************************************************************
 		 * Allocate a new procedure description block
@@ -2014,7 +2033,7 @@ pltcl_process_SPI_result(Tcl_Interp *interp,
  * pltcl_SPI_prepare()		- Builtin support for prepared plans
  *				  The Tcl command SPI_prepare
  *				  always saves the plan using
- *				  SPI_saveplan and returns a key for
+ *				  SPI_keepplan and returns a key for
  *				  access. There is no chance to prepare
  *				  and not save the plan currently.
  **********************************************************************/
@@ -2026,7 +2045,6 @@ pltcl_SPI_prepare(ClientData cdata, Tcl_Interp *interp,
 	int			nargs;
 	CONST84 char **args;
 	pltcl_query_desc *qdesc;
-	void	   *plan;
 	int			i;
 	Tcl_HashEntry *hashent;
 	int			hashnew;
@@ -2105,22 +2123,18 @@ pltcl_SPI_prepare(ClientData cdata, Tcl_Interp *interp,
 		 * Prepare the plan and check for errors
 		 ************************************************************/
 		UTF_BEGIN;
-		plan = SPI_prepare(UTF_U2E(argv[1]), nargs, qdesc->argtypes);
+		qdesc->plan = SPI_prepare(UTF_U2E(argv[1]), nargs, qdesc->argtypes);
 		UTF_END;
 
-		if (plan == NULL)
+		if (qdesc->plan == NULL)
 			elog(ERROR, "SPI_prepare() failed");
 
 		/************************************************************
 		 * Save the plan into permanent memory (right now it's in the
 		 * SPI procCxt, which will go away at function end).
 		 ************************************************************/
-		qdesc->plan = SPI_saveplan(plan);
-		if (qdesc->plan == NULL)
-			elog(ERROR, "SPI_saveplan() failed");
-
-		/* Release the procCxt copy to avoid within-function memory leak */
-		SPI_freeplan(plan);
+		if (SPI_keepplan(qdesc->plan))
+			elog(ERROR, "SPI_keepplan() failed");
 
 		pltcl_subtrans_commit(oldcontext, oldowner);
 	}
@@ -2256,7 +2270,7 @@ pltcl_SPI_execute_plan(ClientData cdata, Tcl_Interp *interp,
 	}
 
 	/************************************************************
-	 * If there was a argtype list on preparation, we need
+	 * If there was an argtype list on preparation, we need
 	 * an argument value list now
 	 ************************************************************/
 	if (qdesc->nargs > 0)
@@ -2409,8 +2423,7 @@ pltcl_set_tuple_values(Tcl_Interp *interp, CONST84 char *arrayname,
 	CONST84 char *nullname = NULL;
 
 	/************************************************************
-	 * Prepare pointers for Tcl_SetVar2() below and in array
-	 * mode set the .tupno element
+	 * Prepare pointers for Tcl_SetVar2() below
 	 ************************************************************/
 	if (arrayname == NULL)
 	{
@@ -2421,6 +2434,12 @@ pltcl_set_tuple_values(Tcl_Interp *interp, CONST84 char *arrayname,
 	{
 		arrptr = &arrayname;
 		nameptr = &attname;
+
+		/*
+		 * When outputting to an array, fill the ".tupno" element with the
+		 * current tuple number.  This will be overridden below if ".tupno" is
+		 * in use as an actual field name in the rowtype.
+		 */
 		snprintf(buf, sizeof(buf), "%d", tupno);
 		Tcl_SetVar2(interp, arrayname, ".tupno", buf, 0);
 	}
